from django.conf import settings

from django_sharding_library.exceptions import NonExistentDatabaseException, ShardedModelInitializationException
from django_sharding_library.fields import ShardedIDFieldMixin


def model_config(shard_group=None, database=None):
    """
    A decorator for marking a model as being either sharded or stored on a
    particular database. When sharding, it does some verification to ensure
    that the model is defined correctly.
    """
    def configure(cls):
        if database and shard_group:
            raise ShardedModelInitializationException('A model cannot be both sharded and stored on a particular database.')

        if not database and not shard_group:
            raise ShardedModelInitializationException('The model should be either sharded or stored on a database in the `model_config` decorator is used.')

        if database:
            if database not in settings.DATABASES or settings.DATABASES[database].get('PRIMARY'):
                raise NonExistentDatabaseException(
                    'Unable to place {} in {} as that is not an existing primary database in the system.'.format(cls._meta.model_name, database)
                )
            setattr(cls, 'django_sharding__database', database)

        if shard_group:
            sharded_fields = filter(lambda field: issubclass(type(field), ShardedIDFieldMixin), cls._meta.fields)
            if not sharded_fields:
                raise ShardedModelInitializationException('All sharded models require a ShardedIDFieldMixin.')

            if not filter(lambda field: field == cls._meta.pk, sharded_fields):
                raise ShardedModelInitializationException('All sharded models require the ShardedAutoIDField to be the primary key. Set primary_key=True on the field.')

            if not callable(getattr(cls, 'get_shard', None)):
                raise ShardedModelInitializationException('You must define a get_shard method on the sharded model.')

            setattr(cls, 'django_sharding__shard_group', shard_group)
            setattr(cls, 'django_sharding__is_sharded', True)

        return cls
    return configure



from mock import patch

from django.conf import settings
from django.db.utils import DataError, IntegrityError
from django.test import TestCase

from django_sharding_library.constants import Backends
from django_sharding_library.fields import ShardedIDFieldMixin, ShardLocalStorageFieldMixin, ShardStorageFieldMixin, ShardForeignKeyStorageFieldMixin, ShardForeignKeyStorageField
from django_sharding_library.id_generation_strategies import BaseIDGenerationStrategy
from tests.models import ShardedModelIDs, ShardedTestModelIDs, TestModel, ShardStorageTable


class BigAutoFieldTestCase(TestCase):
    def test_largest_id(self):
        if settings.DATABASES['default']['ENGINE'] == Backends.POSTGRES:
            max_id = 9223372036854775807
        else:
            max_id = 18446744073709551615

        item = ShardedModelIDs.objects.create(id=max_id, stub=None)
        self.assertEqual(item.pk, max_id)
        try:
            ShardedModelIDs.objects.create(id=max_id + 1, stub=None)
        except (DataError, OverflowError, IntegrityError):
            pass
        else:
            raise Exception('Field was able to hold an integer as large as {}.'.format(max_id + 1))


class FakeIDGenerationStrategy(BaseIDGenerationStrategy):
    def __init__(self):
        self.call_count = 0
        self.id = 0

    def get_next_id(self):
        self.call_count += 1
        self.id += 1
        return self.id


class TableShardedIDFieldTestCase(TestCase):
    def test_largest_id(self):
        if settings.DATABASES['app_shard_001']['ENGINE'] == Backends.POSTGRES:
            max_id = 9223372036854775807
        else:
            max_id = 18446744073709551615

        item = TestModel.objects.using('app_shard_001').create(id=max_id, user_pk=1)
        self.assertEqual(item.pk, max_id)
        try:
            TestModel.objects.using('app_shard_001').create(id=max_id + 1, user_pk=1)
        except (DataError, OverflowError, IntegrityError):
            pass
        else:
            raise Exception('Field was able to hold an integer as large as {}.'.format(max_id + 1))

    def test_get_pk_value_on_save_calls_strategy_get_next_id(self):
        for i in xrange(100):
            instance = TestModel.objects.using('app_shard_001').create(user_pk=1)
            self.assertEqual(ShardedTestModelIDs.objects.using('app_shard_001').latest('pk').pk, instance.pk)


class TableShardedIDMixinTestCase(TestCase):
    def test_initialization_makes_id_generation_available(self):
        field = ShardedIDFieldMixin(strategy=FakeIDGenerationStrategy())
        self.assertTrue(hasattr(field.strategy, 'get_next_id'))

    def test_uses_get_next_id_in_get_pk_value_on_save(self):
        class FakeModel(object):
            pk = None

        field = ShardedIDFieldMixin(strategy=FakeIDGenerationStrategy())
        for i in xrange(1, 100):
            self.assertEqual(field.get_pk_value_on_save(FakeModel()), i)
        self.assertEqual(field.strategy.id, 99)
        self.assertEqual(field.strategy.call_count, 99)


class ShardStorageFieldMixinTestCase(TestCase):
    def test_takes_shard_group_and_sets_attributes(self):
        sut = ShardStorageFieldMixin(shard_group='testing')
        self.assertEqual(sut.django_sharding__shard_group, 'testing')
        self.assertEqual(sut.django_sharding__stores_shard, True)
        self.assertFalse(hasattr(sut, 'django_sharding__use_signal'))


class ShardLocalStorageFieldMixinTestCase(TestCase):
    def test_takes_shard_group_and_sets_attributes(self):
        sut = ShardLocalStorageFieldMixin(shard_group='testing')
        self.assertEqual(sut.django_sharding__shard_group, 'testing')
        self.assertEqual(sut.django_sharding__stores_shard, True)
        self.assertEqual(sut.django_sharding__use_signal, True)


class ShardForeignKeyStorageFieldMixinTestCase(TestCase):
    def test_save_shard_works(self):
        initial_count = ShardStorageTable.objects.count()

        class FakeField(ShardForeignKeyStorageFieldMixin):
            def __init__(self, *args, **kwargs):
                self.name = 'shard_field'
                self.django_sharding__shard_storage_table = ShardStorageTable
                self.django_sharding__shard_group = 'default'

        class FakeModel(object):
            def __init__(self, shard_key):
                self.shard_key = shard_key
                self.shard_field = None

            def get_shard_key(self):
                return self.shard_key

        sut = FakeField(to=ShardStorageTable, shard_group='default')
        model = FakeModel('testing_save_shard')
        sut.save_shard(model)

        latest_entry_in_storage_table = ShardStorageTable.objects.latest('pk')
        self.assertEqual(model.shard_field, latest_entry_in_storage_table)
        self.assertEqual(ShardStorageTable.objects.count(), initial_count + 1)
        self.assertTrue(ShardStorageTable.objects.filter(shard_key='testing_save_shard'))
        self.assertTrue(latest_entry_in_storage_table.shard in settings.DATABASES)

    def test_save_shard_works_when_model_specified_by_string(self):
        initial_count = ShardStorageTable.objects.count()

        class FakeField(ShardForeignKeyStorageFieldMixin):
            def __init__(self, *args, **kwargs):
                self.name = 'shard_field'
                self.django_sharding__shard_storage_table = ShardStorageTable
                self.django_sharding__shard_group = 'default'

        class FakeModel(object):
            def __init__(self, shard_key):
                self.shard_key = shard_key
                self.shard_field = None

            def get_shard_key(self):
                return self.shard_key

        sut = FakeField(to='tests.ShardStorageTable', shard_group='default')
        model = FakeModel('testing_save_shard')
        sut.save_shard(model)

        latest_entry_in_storage_table = ShardStorageTable.objects.latest('pk')
        self.assertEqual(model.shard_field, latest_entry_in_storage_table)
        self.assertEqual(ShardStorageTable.objects.count(), initial_count + 1)
        self.assertTrue(ShardStorageTable.objects.filter(shard_key='testing_save_shard'))
        self.assertTrue(latest_entry_in_storage_table.shard in settings.DATABASES)
    def test_save_shard_works(self):
        initial_count = ShardStorageTable.objects.count()

        class FakeField(ShardForeignKeyStorageFieldMixin):
            def __init__(self, *args, **kwargs):
                self.name = 'shard_field'
                self.django_sharding__shard_storage_table = ShardStorageTable
                self.django_sharding__shard_group = 'default'

        class FakeModel(object):
            def __init__(self, shard_key):
                self.shard_key = shard_key
                self.shard_field = None

            def get_shard_key(self):
                return self.shard_key

        sut = FakeField(to=ShardStorageTable, shard_group='default')
        model = FakeModel('testing_save_shard')
        sut.save_shard(model)

        latest_entry_in_storage_table = ShardStorageTable.objects.latest('pk')
        self.assertEqual(model.shard_field, latest_entry_in_storage_table)
        self.assertEqual(ShardStorageTable.objects.count(), initial_count + 1)
        self.assertTrue(ShardStorageTable.objects.filter(shard_key='testing_save_shard'))
        self.assertTrue(latest_entry_in_storage_table.shard in settings.DATABASES)


class ShardForeignKeyStorageFieldTestCase(TestCase):
    def test_pre_save_calls_save_shard(self):
        sut = ShardForeignKeyStorageField(ShardStorageTable, shard_group='default')
        model_instance = object()
        with patch.object(sut, 'save_shard') as mock_save_shard:
            with self.assertRaises(Exception):
                sut.pre_save(model_instance, False)

        mock_save_shard.assert_called_once_with(model_instance)



from django.apps import apps
from django.db.models import AutoField, CharField, ForeignKey

from django_sharding_library.constants import Backends


class BigAutoField(AutoField):
    """
    An autoincrimenting field which can store an integer from 1 to
    9223372036854775807.
    """
    def db_type(self, connection):
        if connection.settings_dict['ENGINE'] == Backends.MYSQL:
            return 'serial'
        if connection.settings_dict['ENGINE'] == Backends.POSTGRES:
            return 'bigserial'
        return super(BigAutoField, self).db_type(connection)


class ShardedIDFieldMixin(object):
    """
    An autoincrimenting field which takes an id generator class instance
    as an argument and uses the generator to assign each new object a unique
    id.
    Note: This currently must be the primary key of the model and although this
    may be updated in the future, it should not hinder the app to use other
    candidates as unique fields.
    """
    def __init__(self, *args, **kwargs):
        # Remove the strategy from the kwargs so that it doesn't get passed to Django.
        setattr(self, 'strategy', kwargs['strategy'])
        del kwargs['strategy']
        return super(ShardedIDFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ShardedIDFieldMixin, self).deconstruct()

        # Add the strategy from the kwargs so that it does get passed to our model.
        kwargs['strategy'] = getattr(self, 'strategy')
        return name, path, args, kwargs

    def get_pk_value_on_save(self, instance):
        if not instance.pk:
            return self.strategy.get_next_id()
        return instance.pk


class TableShardedIDField(ShardedIDFieldMixin, BigAutoField):
    """
    An autoincrimenting field which takes a `source_table` as an argument in
    order to generate unqiue ids for the sharded model.
    """
    def __init__(self, *args, **kwargs):
        from django_sharding_library.id_generation_strategies import TableStrategy
        kwargs['strategy'] = TableStrategy(backing_model=kwargs['source_table'])
        setattr(self, 'source_table', kwargs['source_table'])
        del kwargs['source_table']
        return super(TableShardedIDField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(TableShardedIDField, self).deconstruct()
        kwargs['source_table'] = getattr(self, 'source_table')
        return name, path, args, kwargs


class ShardedUUID4Field(ShardedIDFieldMixin, CharField):
    def __init__(self, *args, **kwargs):
        from django_sharding_library.id_generation_strategies import UUIDStrategy
        kwargs['strategy'] = UUIDStrategy()
        return super(ShardedUUID4Field, self).__init__(*args, **kwargs)

    def get_pk_value_on_save(self, instance):
        return self.strategy.get_next_id(instance.get_shard())


class ShardStorageFieldMixin(object):
    """
    A mixin for a field used to store a shard for in an instance or parent of an instance.
    """
    def __init__(self, *args, **kwargs):
        setattr(self, 'django_sharding__stores_shard', True)
        setattr(self, 'django_sharding__shard_group', kwargs['shard_group'])
        del kwargs['shard_group']
        return super(ShardStorageFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ShardStorageFieldMixin, self).deconstruct()
        kwargs['shard_group'] = getattr(self, 'django_sharding__shard_group')
        return name, path, args, kwargs


class ShardLocalStorageFieldMixin(ShardStorageFieldMixin):
    """
    The ShardLocalStorageFieldMixin is used for when the shard is stored on the model that
    is being sharded by. i.e. Storing the shard on the User model and sharding by the User.
    """
    def __init__(self, *args, **kwargs):
        setattr(self, 'django_sharding__use_signal', True)
        return super(ShardLocalStorageFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        return super(ShardLocalStorageFieldMixin, self).deconstruct()


class ShardStorageCharField(ShardLocalStorageFieldMixin, CharField):
    """
    A simple char field that stores a shard and uses a signal to generate
    the shard using a pre_save signal.
    """
    pass


class ShardForeignKeyStorageFieldMixin(ShardStorageFieldMixin):
    """
    A mixin for a field used to store a foreign key to another table which
    stores the shard, usually a table which inherits from the ShardStorageModel.
    """
    def __init__(self, *args, **kwargs):
        setattr(self, 'django_sharding__stores_shard', True)
        model_class = kwargs.get('to', args and args[0])
        if type(model_class) == str:
            app_label = model_class.split('.')[0]
            app = apps.get_app_config(app_label)
            model_class = app.get_model(model_class[len(app_label) + 1:])
        setattr(self, 'django_sharding__shard_storage_table', model_class)
        return super(ShardForeignKeyStorageFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ShardForeignKeyStorageFieldMixin, self).deconstruct()
        return name, path, args, kwargs

    def pre_save(self, model_instance, add):
        self.save_shard(model_instance)
        return super(ShardForeignKeyStorageFieldMixin, self).pre_save(model_instance, add)

    def save_shard(self, model_instance):
        shard_key = model_instance.get_shard_key()
        if not getattr(model_instance, self.name):
            shard_storage_table = getattr(self, 'django_sharding__shard_storage_table')
            shard_group = getattr(self, 'django_sharding__shard_group')

            bucketer = apps.get_app_config('django_sharding').get_bucketer(shard_group)
            shard = bucketer.pick_shard(model_instance)
            shard_object, _ = shard_storage_table.objects.get_or_create(shard_key=shard_key)
            if not shard_object.shard:
                shard_object.shard = shard
                shard_object.save()
            setattr(model_instance, self.name, shard_object)


class ShardForeignKeyStorageField(ShardForeignKeyStorageFieldMixin, ForeignKey):
    """
    A simple char field that stores a shard and uses a signal to generate
    the shard using a pre_save signal.
    """
    pass



from django.apps import apps
from django.conf import settings

from django_sharding_library.exceptions import InvalidMigrationException


class ShardedRouter(object):
    """
    A router that is shard-aware and will prevent running migrations on
    the wrong database as well as infer, when possible, which database to
    read or write from.
    """
    def get_shard_group_if_sharded_or_none(self, model):
        if getattr(model, 'django_sharding__is_sharded', False):
            return getattr(model, 'django_sharding__shard_group', None)
        return None

    def get_specific_database_or_none(self, model):
        return getattr(model, 'django_sharding__database', None)

    def get_shard_for_instance(self, instance):
        return instance._state.db or instance.get_shard()

    def get_read_db_routing_strategy(self, shard_group):
        return apps.get_app_config('django_sharding').get_routing_strategy(shard_group)

    def db_for_read(self, model, **hints):
        specific_database = self.get_specific_database_or_none(model)
        if specific_database:
            return specific_database

        if self.get_shard_group_if_sharded_or_none(model):
            instance = hints.get('instance')
            if instance:
                shard = self.get_shard_for_instance(instance)
                # TODO: remove the second, should not use the shard_group attribute anywhere anymore
                shard_group = getattr(model, 'django_sharding__shard_group', getattr(model, 'shard_group', None))
                if not shard_group:
                    raise Exception('Unable to identify the shard_group for the {} model'.format(model))
                routing_strategy = self.get_read_db_routing_strategy(shard_group)
                return routing_strategy.pick_read_db(shard)
        return None

    def db_for_write(self, model, **hints):
        specific_database = self.get_specific_database_or_none(model)
        if specific_database:
            return specific_database

        if self.get_shard_group_if_sharded_or_none(model):
            instance = hints.get('instance')
            if instance:
                db = self.get_shard_for_instance(instance)
                db_config = settings.DATABASES[db]
                return db_config.get('PRIMARY', db)
        return None

    def allow_relation(self, obj1, obj2, **hints):
        """
        Only allow relationships between two items which are both on only one database or
        between sharded items on the same shard.
        """
        if self.get_specific_database_or_none(obj1) != self.get_specific_database_or_none(obj2):
            return False
        elif self.get_specific_database_or_none(obj1):
            return True
        if self.get_shard_group_if_sharded_or_none(type(obj1)) != self.get_shard_group_if_sharded_or_none(type(obj2)):
            return False
        elif self.get_shard_group_if_sharded_or_none(type(obj1)):
            return self.get_shard_for_instance(obj1) == self.get_shard_for_instance(obj2)
        return True

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if settings.DATABASES[db].get('PRIMARY', None):
            return False
        model_name = model_name or hints.get('model_name')
        if not model_name:
            raise InvalidMigrationException(
                'Model name not provided in migration, please pass a `model_name` with the hints passed into the migration.'
            )

        # Sometimes, when extending models from another app i.e. the User Model, the app label
        # is the app label of the app where the change is defined but to app with the model is
        # passed in with the model name.
        try:
            app = apps.get_app_config(app_label)
            model = app.get_model(model_name)
        except LookupError:
            app_label = model_name.split('.')[0]
            app = apps.get_app_config(app_label)
            model = app.get_model(model_name[len(app_label) + 1:])

        single_database = self.get_specific_database_or_none(model)
        shard_group = self.get_shard_group_if_sharded_or_none(model)
        if shard_group and single_database:
            raise InvalidMigrationException(
                'Model marked as both sharded and on a single database, unable to determine where to run migrations for {}.'.format(model_name)
            )
        if single_database:
            return db == single_database
        if shard_group:
            return settings.DATABASES[db]['SHARD_GROUP'] == shard_group
        return db == 'default'



from mock import patch

from django.conf import settings
from django.contrib.auth.models import Group
from django.test import TestCase

from tests.models import TestModel, ShardedTestModelIDs
from django_sharding_library.exceptions import InvalidMigrationException
from django_sharding_library.router import ShardedRouter
from django_sharding_library.routing_read_strategies import BaseRoutingStrategy


class FakeRoutingStrategy(BaseRoutingStrategy):
    """
    A fake router for testing that will always return
    the `testing` DB rather than an existing database.
    """
    def pick_read_db(self, primary_db_name):
        return 'testing'


class RouterReadTestCase(TestCase):

    def setUp(self):
        from django.contrib.auth import get_user_model
        self.sut = ShardedRouter()
        self.user = get_user_model().objects.create_user(username='username', password='pwassword', email='test@example.com')

    def test_sharded_instance_uses_a_routing_read_strategies(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_read(model=TestModel, **hints), 'testing')

    def test_sharded_instance(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        self.assertEqual(self.sut.db_for_read(model=TestModel, **hints), 'app_shard_002')

    def test_sharded_model(self):
        self.assertEqual(self.sut.db_for_read(model=TestModel), None)

    def test_specific_database(self):
        self.assertEqual(self.sut.db_for_read(model=ShardedTestModelIDs), 'app_shard_001')

    def test_specific_database_does_not_use_the_routing_read_strategy(self):
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_read(model=ShardedTestModelIDs), 'app_shard_001')

    def test_other(self):
        from django.contrib.auth import get_user_model
        self.assertEqual(self.sut.db_for_read(model=get_user_model()), None)


class RouterWriteTestCase(TestCase):

    def setUp(self):
        from django.contrib.auth import get_user_model
        self.sut = ShardedRouter()
        self.user = get_user_model().objects.create_user(username='username', password='pwassword', email='test@example.com')

    def test_sharded_instance_does_not_use_a_routing_read_strategies(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_write(model=TestModel, **hints), 'app_shard_002')

    def test_sharded_instance(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        self.assertEqual(self.sut.db_for_write(model=TestModel, **hints), 'app_shard_002')

    def test_sharded_model(self):
        self.assertEqual(self.sut.db_for_write(model=TestModel), None)

    def test_specific_database(self):
        self.assertEqual(self.sut.db_for_write(model=ShardedTestModelIDs), 'app_shard_001')

    def test_specific_database_does_not_use_the_routing_read_strategy(self):
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_write(model=ShardedTestModelIDs), 'app_shard_001')

    def test_other(self):
        from django.contrib.auth import get_user_model
        self.assertEqual(self.sut.db_for_write(model=get_user_model()), None)


class RouterAllowRelationTestCase(TestCase):

    def setUp(self):
        from django.contrib.auth import get_user_model
        self.sut = ShardedRouter()
        self.user = get_user_model().objects.create_user(username='username', password='pwassword', email='test@example.com')

    def test_allow_relation_two_items_on_the_same_non_default_database(self):
        item_one = ShardedTestModelIDs.objects.create(stub=None)
        item_two = ShardedTestModelIDs.objects.create(stub=None)
        self.assertTrue(self.sut.allow_relation(item_one, item_two))

    def test_not_allow_relation_two_items_on_the_different_non_default_database(self):
        class FakeModel(object):
            database = 'one'
        item_one = ShardedTestModelIDs.objects.create(stub=None)

        self.assertFalse(self.sut.allow_relation(item_one, FakeModel()))

    def test_do_not_allow_relation_of_item_on_specific_database__and_non_sharded_non_specific_database_instances(self):
        item_one = ShardedTestModelIDs.objects.create(stub=None)
        self.assertFalse(self.sut.allow_relation(item_one, self.user))

    def test_do_not_allow_relation_of_sharded_models_on_different_shards(self):
        item_one = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        item_two = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        self.assertFalse(self.sut.allow_relation(item_one, item_two))

    def test_allow_relation_of_sharded_models_on_same_shard(self):
        item_one = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        item_two = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        self.assertTrue(self.sut.allow_relation(item_one, item_two))

    def test_do_not_allow_relation_of_sharded_instance_and_item_on_specific_db(self):
        item_one = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        item_two = ShardedTestModelIDs.objects.create(stub=None)
        self.assertFalse(self.sut.allow_relation(item_one, item_two))

    def test_do_not_allow_relation_of_sharded_instance_and_non_sharded_non_specific_database_instances(self):
        item_one = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        self.assertFalse(self.sut.allow_relation(item_one, self.user))

    def test_allow_relation_of_items_on_default_database(self):
        item_one = Group.objects.create(name='test_group')
        self.assertTrue(self.sut.allow_relation(item_one, self.user))


class RouterAllowMigrateTestCase(TestCase):

    def setUp(self):
        self.sut = ShardedRouter()

    def assert_allow_migrate(self, app_label, model_name, can_migrate_default, can_migrate_shard, migratable_db=None):
        self.assertEqual(self.sut.allow_migrate(db='default', app_label=app_label, model_name=model_name), can_migrate_default)
        self.assertEqual(
            all([
                self.sut.allow_migrate(db='app_shard_001_replica_001', app_label=app_label, model_name=model_name),
                self.sut.allow_migrate(db='app_shard_001_replica_002', app_label=app_label, model_name=model_name),
            ]),
            False
        )
        self.assertEqual(
            all([
                self.sut.allow_migrate(db='app_shard_001', app_label=app_label, model_name=model_name),
                self.sut.allow_migrate(db='app_shard_002', app_label=app_label, model_name=model_name),
            ]),
            can_migrate_shard
        )
        if migratable_db:
            self.assertTrue(self.sut.allow_migrate(db=migratable_db, app_label=app_label, model_name=model_name))

    def test_requires_model_name_to_be_passed_in(self):
        with self.assertRaises(InvalidMigrationException):
            self.sut.allow_migrate(db='default', app_label='tests')

        self.sut.allow_migrate(db='default', app_label='tests', model_name='User')
        hints = {'model_name': 'User'}
        self.sut.allow_migrate(db='default', app_label='tests', **hints)

    def test_migrate_replica_will_not_work(self):
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests', model_name='TestModel'))
        self.assertTrue(self.sut.allow_migrate(db='app_shard_001', app_label='tests', model_name='TestModel'))

    def test_migrate_sharded_model_with_specific_database_will_not_work(self):
        sut = ShardedRouter()
        sut.get_specific_database_or_none = lambda self: 'default'
        sut.get_shard_group_if_sharded_or_none = lambda self: 'default'
        with self.assertRaises(InvalidMigrationException):
            sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

        sut.get_specific_database_or_none = lambda self: None
        sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

        sut.get_specific_database_or_none = lambda self: 'default'
        sut.get_shard_group_if_sharded_or_none = lambda self: None
        sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

    def test_django_model_only_allows_on_default(self):
        self.assert_allow_migrate(
            app_label='auth',
            model_name='Group',
            can_migrate_default=True,
            can_migrate_shard=False
        )

    def test_specific_database_only_on_datbase(self):
        self.assert_allow_migrate(
            app_label='tests',
            model_name='ShardedTestModelIDs',
            can_migrate_default=False,
            can_migrate_shard=False,
            migratable_db='app_shard_001'
        )

    def test_specific_sharded_model_only_on_shards(self):
        self.assert_allow_migrate(
            app_label='tests',
            model_name='TestModel',
            can_migrate_default=False,
            can_migrate_shard=True,
        )

    def test_lookup_fallback_if_migration_directory_not_the_same_as_the_model(self):
        self.assert_allow_migrate(
            app_label='tests',
            model_name='auth.User',
            can_migrate_default=True,
            can_migrate_shard=False,
        )
